class MultiTestVars(object):
    '''Expose ansible variabless of a molecule scenario.

    Include ansible facts form molecule host.
    Include roles from ansible project directory.
    Resolve jinja2 template variables.
    '''

    def __init__(
            self,
            moleculeinventory,
            moleculebook):

        # this variable will be returned by the multitestvars fixture
        self._multitestvars = dict()

        for ansiblehost in moleculeinventory.hosts():
            host = str(ansiblehost)
            vars = moleculebook.get_vars(host)
            self._multitestvars[host] = vars

    def get_multitestvars(self):
        return self._multitestvars

    def get_cache(request, cache_key):
        try:
            # read testvars from cache
            # you can enable cache support in molecule.yml:
            # molecule -> verifier -> options
            # option "p: cacheprovider"
            multitestvars = request.config.cache.get(cache_key, None)
        except AttributeError:
            multitestvars = None
        return multitestvars

    def set_cache(self, request, cache_key):
        try:
            request.config.cache.set(cache_key, self._testvars)
        except AttributeError:
            pass



import ansible_runner


class MoleculeInventory(object):

    def __init__(
            self,
            inventory_file):

        out, err = ansible_runner.get_inventory(
            action='list',
            inventories=[str(inventory_file)],
            response_format='json'
        )

        self._hosts = []
        host_groups = []

        try:
            for host_group in out['all']['children']:
                host_groups.append(host_group)
        except (IndexError, KeyError):
            pass

        for host_group in host_groups:
            try:
                for host in out[host_group]['hosts']:
                    self._hosts.append(host)
            except (IndexError, KeyError):
                continue

        if not self._hosts:
            self._hosts = ['localhost']

    def get(self):
        return self._inventory

    def hosts(self):
        return self._hosts



import os
from pathlib import Path
import pytest
from takeltest.moleculeinventory import MoleculeInventory
from takeltest.moleculeenv import MoleculeEnv
from takeltest.moleculebook import MoleculeBook
from takeltest.multitestvars import MultiTestVars
from takeltest.pathlist import PathList


###########################################################
# command line options: testvars group
###########################################################


def pytest_addoption(parser):
    parser.addini('log', 'test')
    testvars_optiongroup = parser.getgroup("testvars")
    testvars_optiongroup.addoption(
        "--testvars-no-gather-roles",
        action="store_false",
        default=True,
        help="do not gather vars from roles")
    testvars_optiongroup.addoption(
        "--testvars-no-extra-vars",
        action="store_false",
        default=True,
        help="do not include extra vars")


###########################################################
# fixtures: testvars option booleans
###########################################################


@pytest.fixture(scope='session')
def gather_roles(request):
    '''testvars option --testvars-no-gather-roles'''
    return request.config.getoption("--testvars-no-gather-roles")


@pytest.fixture(scope='session')
def extra_vars(request):
    '''testvars option --testvars-no-extra-vars'''
    return request.config.getoption("--testvars-no-extra-vars")


###########################################################
# fixtures: testvars option lists and other env vars
###########################################################


@pytest.fixture(scope='session')
def testvars_roles_blocklist():
    '''environment variable TESTVARS_ROLES_BLOCK'''
    try:
        blocklist = os.environ['TESTVARS_ROLES_BLOCK']
    except KeyError:
        return list()
    return blocklist.split(':')


@pytest.fixture(scope='session')
def testvars_roles_exclusivelist(molecule_scenario_directory):
    '''environment variable TESTVARS_ROLES_EXCLUSIVE'''
    try:
        exclusivelist = os.environ['TESTVARS_ROLES_EXCLUSIVE']
    except KeyError:
        return list()
    return exclusivelist.split(':')


@pytest.fixture(scope='session')
def testvars_roles_includelist():
    '''environment variable TESTVARS_ROLES_INCLUDE'''
    try:
        includelist = os.environ['TESTVARS_ROLES_INCLUDE']
    except KeyError:
        return list()
    return includelist.split(':')


@pytest.fixture(scope='session')
def testvars_roles_playbooks():
    '''environment variable TESTVARS_ROLES_PLAYBOOKS'''
    try:
        playbooks = os.environ['TESTVARS_ROLES_PLAYBOOKS']
    except KeyError:
        return None
    return playbooks.split(':')


@pytest.fixture(scope='session')
def testvars_extra_vars(molecule_scenario_directory):
    '''environment variable TESTVARS_EXTRA_VARS'''
    try:
        extra_vars = Path(os.environ['TESTVARS_EXTRA_VARS'])
    except KeyError:
        return list()
    return PathList(
        extra_vars,
        molecule_scenario_directory).get()


###########################################################
# fixtures: molecule inventory
###########################################################


@pytest.fixture(scope='session')
def inventory_file(molecule_ephemeral_directory):
    '''Molecule managed ansible inventory file.'''
    inventory_file = (molecule_ephemeral_directory /
                      'inventory/ansible_inventory.yml')
    inventory_dir = molecule_ephemeral_directory / 'inventory'
    inventory_dir.mkdir(exist_ok=True)
    if not inventory_file.is_file():
        inventory = ("---\n"
                     "all:\n"
                     "  hosts:\n"
                     "    localhost: &id001\n"
                     "      ansible_connection: local\n"
                     "private:\n"
                     "  hosts:\n"
                     "    localhost: *id001")
        inventory_file.write_text(inventory)
    return inventory_file


@pytest.fixture(scope='session')
def moleculeinventory(inventory_file):
    return MoleculeInventory(
        inventory_file)


###########################################################
# fixtures: molecule resources
###########################################################


@pytest.fixture(scope='session')
def molecule_ephemeral_directory(tmp_path_factory):
    '''environment variable MOLECULE_EPHEMERAL_DIRECTORY'''
    try:
        dir = Path(os.environ['MOLECULE_EPHEMERAL_DIRECTORY'])
    except KeyError:
        dir = tmp_path_factory.mktemp('molecule_ephemeral_directory')
    return dir


@pytest.fixture(scope='session')
def molecule_scenario_directory(tmp_path_factory):
    '''environment variable MOLECULE_SCENARIO_DIRECTORY'''
    try:
        dir = Path(os.environ['MOLECULE_SCENARIO_DIRECTORY'])
    except KeyError:
        dir = tmp_path_factory.mktemp('molecule_scenario_directory')
    return dir


@pytest.fixture(scope='session')
def moleculeenv(
        molecule_ephemeral_directory,
        molecule_scenario_directory,
        gather_roles,
        testvars_roles_blocklist,
        testvars_roles_exclusivelist,
        testvars_roles_includelist,
        testvars_roles_playbooks):
    return MoleculeEnv(
        molecule_ephemeral_directory,
        molecule_scenario_directory,
        gather_roles,
        testvars_roles_blocklist,
        testvars_roles_exclusivelist,
        testvars_roles_includelist,
        testvars_roles_playbooks)


###########################################################
# fixtures: ansible runner python api
###########################################################


@pytest.fixture(scope='session')
def moleculebook(
        testvars_extra_vars,
        moleculeinventory,
        moleculeenv):
    '''Run an ansible playbook against a molecule host.'''
    return MoleculeBook(
        testvars_extra_vars,
        moleculeinventory,
        moleculeenv)


###########################################################
# fixtures: testvars helpers
###########################################################


@pytest.fixture(scope='session')
def cache_key(molecule_ephemeral_directory):
    # molecule_ephemeral_directory should be unique for each scenario
    return 'multitestvars' + str(molecule_ephemeral_directory)


###########################################################
# fixtures: molecule/testinfra/ansible helpers
###########################################################


@pytest.fixture(scope='session')
def multitestvars(
        request,
        moleculeinventory,
        moleculebook,
        cache_key):
    '''Expose ansible variables and facts of a molecule test scenario.'''
    multitestvars = MultiTestVars.get_cache(request, cache_key)
    if multitestvars is None:
        multitestvars_object = MultiTestVars(
            moleculeinventory,
            moleculebook)
        multitestvars = multitestvars_object.get_multitestvars()
        multitestvars_object.set_cache(request, cache_key)
    return multitestvars


@pytest.fixture(scope='session')
def testvars(multitestvars):
    '''Expose ansible variables and facts of a molecule test scenario.'''
    return multitestvars[next(iter(multitestvars))]



import json
from pathlib import Path
from ruamel.yaml import YAML
from ruamel.yaml.scanner import ScannerError
import subprocess
import yaml


class MoleculeEnv(object):

    def __init__(
            self,
            molecule_ephemeral_directory,
            molecule_scenario_directory,
            gather_roles,
            testvars_roles_blocklist,
            testvars_roles_exclusivelist,
            testvars_roles_includelist,
            testvars_roles_playbooks):
        self._molecule_ephemeral_directory = molecule_ephemeral_directory
        self._molecule_scenario_directory = molecule_scenario_directory
        self._gather_roles = gather_roles
        self._testvars_roles_blocklist = testvars_roles_blocklist
        self._testvars_roles_exclusivelist = testvars_roles_exclusivelist
        self._testvars_roles_includelist = testvars_roles_includelist
        self._testvars_roles_playbooks = testvars_roles_playbooks
        self._configure_roles_()

    def get_molecule_ephemeral_directory(self):
        return self._molecule_ephemeral_directory

    def get_molecule_scenario_directory(self):
        return self._molecule_scenario_directory

    def get_project_dir(self):
        '''Return ansible project dir.'''

        # start the ansible scenario directory
        path = self.get_molecule_scenario_directory()

        # go up until you find the roles dir
        while path != Path('/'):
            path = path.parent
            if path / 'roles' in [d for d in path.iterdir() if d.is_dir()]:
                return path

        # else return None
        return None

    def get_roles(self):
        '''Return roles as list of pathlib.Path objects'''
        roles_dir = self.get_molecule_ephemeral_directory() / 'roles'

        if not roles_dir.exists():
            return list()

        roles = sorted([d.name for d in roles_dir.iterdir() if d.is_dir()])
        return roles

    def write_playbook(
            self,
            playbook):
        '''Write playbook to disk'''
        project_dir = self._molecule_ephemeral_directory / 'project'
        project_dir.mkdir(exist_ok=True)
        playbook_file = self._molecule_ephemeral_directory / 'project/site.json'
        playbook_file.write_text(json.dumps(playbook))

    def write_extravars(self, vars):
        # extravars = "---\ncurl_my_var: 'curl_my_value'\ncurl_my_ref: \"{{ curl_my_var }}\""
        envdir = self._molecule_ephemeral_directory / 'env'
        envdir.mkdir(exist_ok=True)
        extravars = self._molecule_ephemeral_directory / 'env/extravars'
        extravars.write_text("---\n" + yaml.dump(vars))

    def _configure_roles_(self):
        '''Create symlinks to roles'''
        if self._gather_roles is False:
            return

        roles = None

        # exclusivelist selects roles directly
        if self._get_testvars_roles_exclusivelist_():
            roles = self._get_testvars_roles_exclusivelist_()

        # try to read roles from custom playbooks
        if roles is None:
            playbook_files = self._get_testvars_roles_playbooks_()
            if playbook_files is not None:
                for playbook_file in playbook_files:
                    if roles is None:
                        roles = []
                    roles += self._read_roles_from_playbook_(playbook_file)

        # try to read roles from custom molecule converge playbook
        if roles is None:
            playbook_file = self._read_playbook_file_from_molecule_yml_()
            if playbook_file is not None:
                roles = self._read_roles_from_playbook_(playbook_file)

        # try to read roles from default molecule 3 converge playbook
        if roles is None:
            roles = self._read_roles_from_playbook_('converge.yml')

        # try to read roles from default molecule 2 converge playbook
        if roles is None:
            roles = self._read_roles_from_playbook_('playbook.yml')

        # fallback: select all roles
        if roles is None:
            msd = self.get_molecule_scenario_directory()

            # roles dir viewed from a project molecule scenario
            roles_dir = msd / '../../roles'
            if roles_dir.is_dir():
                roles = sorted([d.name for d in roles_dir.iterdir()
                                if d.is_dir()])
            else:
                # roles dir viewed from a role molecule scenario
                roles_dir = msd / '../../../../roles'
                if roles_dir.is_dir():
                    roles = sorted([d.name for d in roles_dir.iterdir()
                                    if d.is_dir()])

        # apply include
        # by now, roles is a list
        roles = self._roles_apply_includelist_(roles)

        # if roles have been selected
        # then apply blocklist
        # then create symlinks
        # then return
        if roles is not None:
            roles = self._roles_apply_blocklist_(roles)
            self._create_roles_symlinks_(roles)
            return

    def _create_roles_symlinks_(self, roles):
        (self.get_molecule_ephemeral_directory() / 'roles').mkdir(
            exist_ok=True)
        for role in roles:
            self._create_symlink_('roles/' + role)
        (self.get_molecule_ephemeral_directory() / 'project').mkdir(
            exist_ok=True)
        source = self.get_molecule_ephemeral_directory() / 'roles'
        target = self.get_molecule_ephemeral_directory() / 'project/roles'
        try:
            target.symlink_to(source)
        except FileExistsError:
            pass

    def _create_symlink_(self, path):
        '''Create symlink from molecule ephemeral dir to project dir.'''
        project_dir = self.get_project_dir()
        if project_dir is None:
            return
        source = project_dir / path
        target = self.get_molecule_ephemeral_directory() / path
        try:
            target.symlink_to(source)
        except FileExistsError:
            pass

    def _get_molecule_vars_config_(self):
        molecule_vars_config = ''
        vars_config = list()

        group_vars = \
            self._molecule_ephemeral_directory / 'inventory/group_vars'
        host_vars = \
            self._molecule_ephemeral_directory / 'inventory/host_vars'

        if group_vars.is_symlink():
            target = self._molecule_scenario_directory / 'group_vars'
            vars_config.append('Using group_vars symlink to ' + str(target))

        if host_vars.is_symlink():
            target = self._molecule_scenario_directory / 'host_vars'
            vars_config.append('Using host_vars symlink to ' + str(target))

        if vars_config:
            molecule_vars_config = '\n'.join(vars_config)

            return molecule_vars_config

        # else no vars_config so we use the variables defined in molecule.yml
        molecule_yml = self._molecule_scenario_directory / 'molecule.yml'
        molecule_vars_config += \
            'Using variables defined in ' + str(molecule_yml)

        return molecule_vars_config

    def _get_testvars_roles_blocklist_(self):
        return self._testvars_roles_blocklist

    def _get_testvars_roles_exclusivelist_(self):
        return self._testvars_roles_exclusivelist

    def _get_testvars_roles_includelist_(self):
        return self._testvars_roles_includelist

    def _get_testvars_roles_playbooks_(self):
        return self._testvars_roles_playbooks

    def _read_playbook_file_from_molecule_yml_(self):
        molecule_yml_path = self.get_molecule_scenario_directory() / \
                            'molecule.yml'
        yaml = YAML(typ='safe')
        try:
            molecule_yml = yaml.load(molecule_yml_path)
            playbook_file = \
                molecule_yml['provisioner']['playbooks']['converge']
            command = [
                'bash',
                '-c',
                f"echo {playbook_file}"
            ]
            result = subprocess.run(
                command,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE).stdout.decode('UTF-8').strip()
            playbook_file = Path(result)
            return playbook_file
        except (FileNotFoundError, ScannerError, KeyError):
            return None

    def _read_roles_from_playbook_(self, playbook_file):
        playbook_path = self.get_molecule_scenario_directory() / playbook_file
        if not playbook_path.is_file():
            return None

        yaml = YAML(typ='safe')
        try:
            playbook = yaml.load(playbook_path)
            roles = playbook[0]['roles']
        except (FileNotFoundError, ScannerError, KeyError):
            return None

        # Roles are either a dict with string val, or string
        # this normalises them to list of strings
        roles = [r['role'] if isinstance(r, dict) else r for r in roles]

        return roles

    def _roles_apply_blocklist_(self, roles):
        roles_not_blocklisted = list()
        for role in roles:
            if role not in self._get_testvars_roles_blocklist_():
                roles_not_blocklisted.append(role)
        return roles_not_blocklisted

    def _roles_apply_includelist_(self, roles):
        roles_include = self._get_testvars_roles_includelist_()
        for role in roles_include:
            if role not in roles:
                roles.append(role)
        return roles



import ansible_runner


class MoleculeBook(object):
    '''Run an ansible playbook against a molecule host'''

    def __init__(self,
                 testvars_extra_vars,
                 moleculeinventory,
                 moleculeenv):
        self._hosts = moleculeinventory.hosts()
        self._moleculeenv = moleculeenv
        self._testvars_extra_vars = testvars_extra_vars
        self._playbook = dict()
        self.create()

    def get(self):
        '''Get the ansible playbook'''
        return self._playbook

    def set(self, playbook):
        '''Set an ansible playbook'''
        self._playbook = playbook

    def get_molecule_scenario_directory(self):
        return self._moleculeenv.get_molecule_scenario_directory()

    def get_roles(self):
        return self._moleculeenv.get_roles()

    def create(self,
               gather_facts=False,
               gather_roles=True,
               extra_vars=True,
               host=None):
        '''Create an ansible playbook using the ansible python api.'''
        if host is None:
            host = next(iter(self._hosts))

        play = dict(
            name="ansible playbook",
            hosts=host,
            gather_facts=str(gather_facts),
            vars_files=list(),
            roles=list(),
            tasks=list(),
        )
        playbook = [play]

        # include extra vars files
        if extra_vars:
            for path in self._get_extra_vars_():
                playbook[0]['vars_files'].append(str(path))

        # include roles
        if gather_roles:
            for role in self.get_roles():
                playbook[0]['roles'].append({'name': role, 'when': 'False'})
        self._playbook = playbook

    def add_task_debug_msg(self, msg, playbook, name='Debug'):
        '''Add a task using the ansible debug module'''
        task = {'name': name,
                'ansible.builtin.debug': {'msg': msg}}
        playbook[0]['tasks'].append(task)
        return playbook

    def add_task_get_vars(self, playbook, name='Get vars'):
        '''Add a task using the ansible debug module to gather variables'''
        msg = "{{ vars }}"
        task = {'name': name,
                'ansible.builtin.debug': {
                    'msg': msg}}
        playbook[0]['tasks'].append(task)
        return playbook

    def add_task_include_vars_dir(
            self,
            vars_dir,
            playbook,
            name='Include vars'):
        '''Add a task using the ansible include_vars module'''
        task = {'name': name,
                'ansible.builtin.include_vars': {
                    'dir': vars_dir}}
        playbook[0]['tasks'].append(task)
        return playbook

    def get_vars(
            self,
            host=None):
        '''Return ansible facts and vars of a molecule host.

        Args:
            gather_facts (bool): gather ansible_facts from a molecule host?
                Defaults to True.
                A playbook will be run with ``gather_facts:true``.
            extra_vars (bool): include extra vars from vars files?
                Defaults to True.
                An include_vars task will be added to include extra vars files
                specified in the environment variable TESTVARS_EXTRA_VARS
            host (string)

        Returns:
            dict: resolved ansible variables and facts
        '''

        playbook = self._playbook
        playbook[0]['hosts'] = host
        playbook[0]['gather_facts'] = True
        task_name = 'moleculebook_get_vars'

        # this is where the magic happens:
        # calling the debug module this way yields all variables
        self.add_task_get_vars(playbook, task_name)

        runner = self.run(playbook)
        vars = self._get_vars_from_event_data(runner, host, task_name)

        # some (role) variables may not be resolved yet
        # so we write all vars to env/extravars
        # (where they do get resolved)
        # and run ansible again
        self._moleculeenv.write_extravars(vars)
        runner = self.run(playbook)
        vars = self._get_vars_from_event_data(runner, host, task_name)

        # we write the resolved vars to disk again
        # you can inspect them at ~|.cache/molecule/.../env/extravars
        self._moleculeenv.write_extravars(vars)

        return vars

    def run(self, playbook):
        '''Run the ansible playbook'''
        self._moleculeenv.write_playbook(playbook)
        private_data_dir = self._moleculeenv.get_molecule_ephemeral_directory()
        r = ansible_runner.run(
            private_data_dir=private_data_dir,
            playbook='site.json')
        return r

    def _get_extra_vars_(self):
        return self._testvars_extra_vars

    def _get_molecule_scenario_directory_(self):
        return self._get_molecule_scenario_directory()

    def _get_vars_from_event_data(self, runner, host, task_name):
        for event in runner.events:
            try:
                if event['event'] == 'runner_on_ok':
                    if event['event_data']['task'] == task_name:
                        # Merge and overwrite all variables
                        # (some still unresolved)
                        # with the (resolved) hostvars
                        return (
                                event['event_data']['res']['msg'] |
                                event['event_data']['res']['msg']['hostvars'][host]
                        )
            except (IndexError, KeyError):
                continue
        return dict()



from takeltest.boilerplate import hosts  # noqa F401
import takeltest.moleculebook
import takeltest.moleculeenv
import takeltest.plugin
import takeltest.multitestvars  # noqa F401

name = "takeltest"
